# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import unittest

from odcs.server import db
from sqlalchemy import event
from odcs.server.events import cache_composes_if_state_changed
from odcs.server.events import start_to_publish_messages

from flask_sqlalchemy import SignallingSession
from mock import patch


class AnyStringWith(str):
    def __eq__(self, other):
        return self in str(other)


class ConfigPatcher(object):
    def __init__(self, config_obj):
        self.objects = []
        self.config_obj = config_obj

    def patch(self, key, value):
        try:
            obj = patch.object(self.config_obj, key, new=value)
        except Exception:
            self.stop()
            raise
        self.objects.append(obj)

    def start(self):
        for obj in self.objects:
            obj.start()

    def stop(self):
        for obj in self.objects:
            obj.stop()


class ModelsBaseTest(unittest.TestCase):
    """Base test case for models

    Database and schemas are initialized on behalf of developers.
    """

    disable_event_handlers = True

    def setUp(self):
        # Not all tests need handlers of event after_flush and after_commit.
        if event.contains(
            SignallingSession, "after_flush", cache_composes_if_state_changed
        ):
            event.remove(
                SignallingSession, "after_flush", cache_composes_if_state_changed
            )
        if event.contains(SignallingSession, "after_commit", start_to_publish_messages):
            event.remove(SignallingSession, "after_commit", start_to_publish_messages)

        db.session.remove()
        db.drop_all()
        db.create_all()
        db.session.commit()

        setup_composes = getattr(self, "setup_composes", None)
        if setup_composes is not None:
            assert callable(setup_composes)
            setup_composes()

        if hasattr(self, "setup_composes"):
            getattr(self, "setup_composes")()

        # And, if tests which need such event handlers or just tests those
        # handlers, add them back.
        if not self.disable_event_handlers:
            event.listen(
                SignallingSession, "after_flush", cache_composes_if_state_changed
            )
            event.listen(SignallingSession, "after_commit", start_to_publish_messages)

    def tearDown(self):
        if not self.disable_event_handlers:
            event.remove(
                SignallingSession, "after_flush", cache_composes_if_state_changed
            )
            event.remove(SignallingSession, "after_commit", start_to_publish_messages)

        db.session.remove()
        db.drop_all()
        db.session.commit()

        # Nothing special here. Just do what should be done in tearDown to
        # to restore enviornment for each test method.
        event.listen(SignallingSession, "after_flush", cache_composes_if_state_changed)
        event.listen(SignallingSession, "after_commit", start_to_publish_messages)



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
from logging import getLogger

from odcs.server import conf

log = getLogger(__name__)

__all__ = ("publish",)


def publish(msgs):
    """Start to send messages to message broker"""
    backend = _get_messaging_backend()
    if backend is not None:
        backend(msgs)


def _umb_send_msg(msgs):
    """Send message to Unified Message Bus"""

    import proton
    from rhmsg.activemq.producer import AMQProducer

    config = {
        "urls": conf.messaging_broker_urls,
        "certificate": conf.messaging_cert_file,
        "private_key": conf.messaging_key_file,
        "trusted_certificates": conf.messaging_ca_cert,
    }
    with AMQProducer(**config) as producer:
        producer.through_topic(conf.messaging_topic)

        for msg in msgs:
            outgoing_msg = proton.Message()
            outgoing_msg.body = json.dumps(msg)
            producer.send(outgoing_msg)


def _fedora_messaging_send_msg(msgs):
    """Send message to fedora-messaging."""
    from fedora_messaging import api, config

    config.conf.setup_logging()

    for msg in msgs:
        # "event" is typically just "state-changed"
        event = msg.get("event", "event")
        topic = "odcs.compose.%s" % event

        api.publish(api.Message(topic=topic, body=msg))


def _get_messaging_backend():
    if conf.messaging_backend == "rhmsg":
        return _umb_send_msg
    elif conf.messaging_backend == "fedora-messaging":
        return _fedora_messaging_send_msg
    elif conf.messaging_backend:
        raise ValueError("Unknown messaging backend {0}".format(conf.messaging_backend))
    else:
        return None



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import six
import unittest

from mock import patch
import time

from odcs.server.utils import execute_cmd


class TestUtilsExecuteCmd(unittest.TestCase):
    def setUp(self):
        super(TestUtilsExecuteCmd, self).setUp()

    def tearDown(self):
        super(TestUtilsExecuteCmd, self).tearDown()

    def test_execute_cmd_timeout_called(self):
        start_time = time.time()
        with six.assertRaisesRegex(self, RuntimeError, "Compose has taken more time.*"):
            execute_cmd(["/usr/bin/sleep", "5"], timeout=1)
        stop_time = time.time()

        self.assertTrue(stop_time - start_time < 2)

    @patch("odcs.server.utils._kill_process_group")
    def test_execute_cmd_timeout_not_called(self, killpg):
        execute_cmd(["/usr/bin/true"], timeout=1)
        time.sleep(2)
        killpg.assert_not_called()



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import os
from datetime import datetime
from datetime import timedelta

from odcs.server import db, conf
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_RESULTS, COMPOSE_STATES
from odcs.server.models import User
from odcs.server.pungi import PungiSourceType

import pytest

from .utils import ModelsBaseTest


class TestModels(ModelsBaseTest):
    def test_creating_event_and_builds(self):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()
        db.session.expire_all()

        c = db.session.query(Compose).filter(compose.id == 1).one()
        c.pungi_config_dump = "test"
        self.assertEqual(c.owner, "me")
        self.assertEqual(c.source_type, PungiSourceType.MODULE)
        self.assertEqual(c.source, "testmodule-master")
        self.assertEqual(c.results, COMPOSE_RESULTS["repository"])
        self.assertTrue(c.time_to_expire)

        expected_json = {
            "source_type": 2,
            "state": 0,
            "time_done": None,
            "state_name": "wait",
            "state_reason": None,
            "source": u"testmodule-master",
            "owner": u"me",
            "result_repo": "http://localhost/odcs/odcs-1/compose/Temporary",
            "result_repofile": "http://localhost/odcs/odcs-1/compose/Temporary/odcs-1.repo",
            "time_submitted": c.json()["time_submitted"],
            "id": 1,
            "time_started": None,
            "time_removed": None,
            "removed_by": None,
            "time_to_expire": c.json()["time_to_expire"],
            "flags": [],
            "results": ["repository"],
            "sigkeys": "",
            "koji_event": None,
            "koji_task_id": None,
            "packages": None,
            "builds": None,
            "arches": "x86_64",
            "multilib_arches": "",
            "multilib_method": 0,
            "lookaside_repos": None,
            "modular_koji_tags": None,
            "module_defaults_url": None,
            "label": None,
            "compose_type": None,
            "pungi_compose_id": None,
            "pungi_config_dump": "test",
            "target_dir": "default",
            "toplevel_url": "http://localhost/odcs/odcs-1",
            "scratch_modules": None,
            "parent_pungi_compose_ids": None,
            "scratch_build_tasks": None,
        }
        self.assertEqual(c.json(True), expected_json)

    def test_compose_paths(self):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        compose.id = 1
        self.assertEqual(compose.toplevel_dir, os.path.join(conf.target_dir, "odcs-1"))
        self.assertEqual(
            compose.result_repofile_path,
            os.path.join(conf.target_dir, "odcs-1/compose/Temporary/odcs-1.repo"),
        )
        self.assertEqual(
            compose.result_repo_dir,
            os.path.join(conf.target_dir, "odcs-1/compose/Temporary"),
        )

    def test_target_dir_none(self):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        compose.target_dir = None
        db.session.commit()
        self.assertEqual(compose.target_dir, conf.target_dir)

    def test_create_copy(self):
        """
        Tests that the Compose atttributes stored in database are copied
        by Compose.create_copy() method.
        """
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        # Generate non-default data for every attribute in compose, so we can
        # later verify they have been copied to new compose.
        for c in Compose.__table__.columns:
            t = str(c.type)
            if t == "INTEGER":
                new_value = 4
            elif t == "VARCHAR":
                new_value = "non default value"
            elif t == "DATETIME":
                new_value = datetime.utcnow()
            else:
                raise ValueError(
                    "New column type %r added, please handle it " "in this test" % t
                )
            setattr(compose, c.name, new_value)

        db.session.commit()
        db.session.expire_all()

        copy = Compose.create_copy(db.session, compose)
        for c in Compose.__table__.columns:
            # Following are list of fields which should not be copied
            # in create_copy() method.
            if c.name in [
                "id",
                "state",
                "state_reason",
                "time_to_expire",
                "time_done",
                "time_submitted",
                "time_removed",
                "removed_by",
                "reused_id",
                "koji_task_id",
                "time_started",
                "pungi_compose_id",
                "celery_task_id",
            ]:
                assertMethod = self.assertNotEqual
            else:
                assertMethod = self.assertEqual
            assertMethod(
                [c.name, getattr(compose, c.name)], [c.name, getattr(copy, c.name)]
            )


class TestUserModel(ModelsBaseTest):
    def test_find_by_email(self):
        db.session.add(User(username="tester1"))
        db.session.add(User(username="admin"))
        db.session.commit()

        user = User.find_user_by_name("admin")
        self.assertEqual("admin", user.username)

    def test_create_user(self):
        User.create_user(username="tester2")
        db.session.commit()

        user = User.find_user_by_name("tester2")
        self.assertEqual("tester2", user.username)

    def test_no_group_is_added_if_no_groups(self):
        User.create_user(username="tester1")
        db.session.commit()

        user = User.find_user_by_name("tester1")
        self.assertEqual("tester1", user.username)


class ComposeModel(ModelsBaseTest):
    """Test Compose Model"""

    def setUp(self):
        super(ComposeModel, self).setUp()

        self.c1 = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
        )
        self.c2 = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1",
        )
        self.c3 = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1",
        )
        self.c4 = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1",
        )

        map(db.session.add, (self.c1, self.c2, self.c3, self.c4))
        db.session.commit()

        self.c1.reused_id = self.c3.id
        self.c2.reused_id = self.c3.id
        self.c3.reused_id = self.c4.id
        db.session.commit()

    def test_get_reused_compose(self):
        compose = self.c3.get_reused_compose()
        self.assertEqual(self.c4, compose)

    def test_get_reusing_composes(self):
        composes = self.c3.get_reusing_composes()
        self.assertEqual(2, len(composes))
        self.assertEqual([self.c1, self.c2], list(composes))

    def test_extend_expiration(self):
        from_now = datetime.utcnow()
        seconds_to_live = 100
        self.c1.extend_expiration(from_now, seconds_to_live)
        db.session.commit()

        expected_time_to_expire = from_now + timedelta(seconds=seconds_to_live)
        self.assertEqual(expected_time_to_expire, self.c1.time_to_expire)

    def test_not_extend_expiration(self):
        from_now = datetime.utcnow()
        seconds_to_live = (self.c1.time_to_expire - datetime.utcnow()).seconds / 2

        orig_time_to_expire = self.c1.time_to_expire
        self.c1.extend_expiration(from_now, seconds_to_live)

        self.assertEqual(orig_time_to_expire, self.c1.time_to_expire)

    def test_composes_to_expire(self):
        now = datetime.utcnow()
        self.c1.time_to_expire = now - timedelta(seconds=60)
        self.c1.state = COMPOSE_STATES["done"]
        self.c2.time_to_expire = now - timedelta(seconds=60)
        self.c2.state = COMPOSE_STATES["failed"]
        self.c3.time_to_expire = now + timedelta(seconds=60)
        self.c3.state = COMPOSE_STATES["done"]
        self.c4.time_to_expire = now + timedelta(seconds=60)
        self.c4.state = COMPOSE_STATES["failed"]
        db.session.commit()

        composes = Compose.composes_to_expire()
        self.assertTrue(self.c1 in composes)
        self.assertTrue(self.c2 in composes)
        self.assertTrue(self.c3 not in composes)

    def test_transition_to_done_updates_time_to_expire(self):
        in_five_minutes = datetime.utcnow() + timedelta(seconds=300)
        self.c1.transition(COMPOSE_STATES["done"], "it's finished", in_five_minutes)
        # The compose should expire about 60 seconds (give or take a tenth of a
        # second) after the time it was finished.
        expires_in = self.c1.time_to_expire - self.c1.time_done
        assert expires_in.total_seconds() == pytest.approx(60, abs=0.1)

    def test_transition_to_failed_updates_time_to_expire(self):
        now = datetime.utcnow()
        in_five_minutes = now + timedelta(seconds=300)
        # Fail the compose in five minutes...
        self.c1.transition(COMPOSE_STATES["failed"], "it's finished", in_five_minutes)
        # so that it expires in about 6 minutes
        expires_in = self.c1.time_to_expire - now
        assert expires_in.total_seconds() == pytest.approx(360, abs=0.1)

    def test_transition_to_generating_updates_time_started(self):
        now = datetime.utcnow()
        in_five_minutes = now + timedelta(seconds=300)
        self.c1.transition(COMPOSE_STATES["generating"], "it started", in_five_minutes)
        assert self.c1.time_started == in_five_minutes



"""Add index on Compose.source_type and Compose.state

Revision ID: c370b90de998
Revises: f24a36cc8a16
Create Date: 2017-09-21 11:47:09.381048

"""

# revision identifiers, used by Alembic.
revision = "c370b90de998"
down_revision = "f24a36cc8a16"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index("idx_source_type__state", "composes", ["source_type", "state"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("idx_source_type__state", table_name="composes")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from logging import getLogger

from flask import Flask, jsonify
from flask_login import LoginManager
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import BadRequest, Unauthorized

from odcs.server.logger import init_logging
from odcs.server.config import init_config
from odcs.server.proxy import ReverseProxy
from odcs.server.errors import NotFound, Forbidden

import pkg_resources

try:
    version = pkg_resources.get_distribution("odcs").version
except pkg_resources.DistributionNotFound:
    version = "unknown"

app = Flask(__name__)
app.wsgi_app = ReverseProxy(app.wsgi_app)

conf = init_config(app)

db = SQLAlchemy(app)

init_logging(conf)
log = getLogger(__name__)

login_manager = LoginManager()
login_manager.init_app(app)

from odcs.server import views  # noqa

from odcs.server.auth import init_auth  # noqa

init_auth(login_manager, conf.auth_backend)


def json_error(status, error, message):
    response = jsonify({"status": status, "error": error, "message": message})
    response.status_code = status
    return response


@app.errorhandler(NotFound)
def notfound_error(e):
    """Flask error handler for NotFound exceptions"""
    return json_error(404, "Not Found", e.args[0])


@app.errorhandler(Unauthorized)
def unauthorized_error(e):
    """Flask error handler for Unauthorized exceptions"""
    return json_error(401, "Unauthorized", e.description)


@app.errorhandler(Forbidden)
def forbidden_error(e):
    """Flask error handler for Forbidden exceptions"""
    return json_error(403, "Forbidden", e.args[0])


@app.errorhandler(BadRequest)
def badrequest_error(e):
    """Flask error handler for RuntimeError exceptions"""
    return json_error(400, "Bad Request", e.get_description())


@app.errorhandler(ValueError)
def validationerror_error(e):
    """Flask error handler for ValueError exceptions"""
    return json_error(400, "Bad Request", str(e))


@app.errorhandler(Exception)
def internal_server_error(e):
    """Flask error handler for RuntimeError exceptions"""
    log.exception("Internal server error: %s", e)
    return json_error(500, "Internal Server Error", str(e))



# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import time
import os
from mock import patch, MagicMock, call

from .utils import ModelsBaseTest
import odcs.server
from odcs.server import db, conf
from odcs.server.cache import KojiTagCache
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_RESULTS, COMPOSE_STATES, PungiSourceType


class TestKojiTagCache(ModelsBaseTest):
    def setUp(self):
        super(TestKojiTagCache, self).setUp()
        compose = MagicMock()
        compose.target_dir = conf.target_dir
        self.cache = KojiTagCache(compose)

    def test_cached_compose_dir(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        ret = self.cache.cached_compose_dir(c)
        expected = os.path.join(conf.target_dir, "koji_tag_cache/f26-0--x86_64")
        self.assertEqual(ret, expected)

    def test_compose_dir_flags(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=1,
        )
        ret = self.cache.cached_compose_dir(c)
        expected = os.path.join(conf.target_dir, "koji_tag_cache/f26-1--x86_64")
        self.assertEqual(ret, expected)

    def test_compose_dir_sigkeys(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=1,
            sigkeys="a b c",
        )
        ret = self.cache.cached_compose_dir(c)
        expected = os.path.join(conf.target_dir, "koji_tag_cache/f26-1-a-b-c-x86_64")
        self.assertEqual(ret, expected)

    def test_compose_dir_sigkeys_arches_unsorted(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=1,
            sigkeys="c b a",
            arches="i686 x86_64",
        )
        ret = self.cache.cached_compose_dir(c)
        expected = os.path.join(
            conf.target_dir, "koji_tag_cache/f26-1-a-b-c-i686-x86_64"
        )
        self.assertEqual(ret, expected)

    @patch("os.path.exists")
    def test_is_cached(self, exists):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )

        for cached in [True, False]:
            exists.return_value = cached
            ret = self.cache.is_cached(c)
            self.assertEqual(ret, cached)
            exists.assert_called_once_with(
                os.path.join(conf.target_dir, "koji_tag_cache/f26-0--x86_64")
            )
            exists.reset_mock()

    @patch("shutil.copytree")
    def test_reuse_cached(self, copytree):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()
        self.cache.reuse_cached(c)
        copytree.assert_called_once_with(
            os.path.join(conf.target_dir, "koji_tag_cache/f26-0--x86_64"),
            os.path.join(conf.target_dir, "koji_tag_cache/odcs-1-1-19700101.n.0"),
            symlinks=True,
        )

    @patch("shutil.copytree")
    @patch("os.path.realpath")
    @patch("shutil.rmtree")
    def test_update_cache(self, rmtree, realpath, copytree):
        realpath.return_value = "/tmp/real/path"
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()
        for done in [True, False]:
            c.state = COMPOSE_STATES["done"] if done else COMPOSE_STATES["failed"]
            self.cache.update_cache(c)
            if done:
                copytree.assert_called_once_with(
                    "/tmp/real/path",
                    os.path.join(conf.target_dir, "koji_tag_cache/f26-0--x86_64"),
                    symlinks=True,
                )
            else:
                copytree.assert_not_called()
            copytree.reset_mock()
            rmtree.assert_not_called()

    @patch("shutil.copytree")
    @patch("os.path.realpath")
    @patch("shutil.rmtree")
    @patch("os.path.exists")
    def test_update_cache_rmtree_if_exists(self, exists, rmtree, realpath, copytree):
        realpath.return_value = "/tmp/real/path"
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        c.state = COMPOSE_STATES["done"]
        db.session.add(c)
        db.session.commit()

        self.cache.update_cache(c)
        rmtree.assert_called_once_with(
            os.path.join(conf.target_dir, "koji_tag_cache/f26-0--x86_64")
        )

    @patch("os.listdir")
    @patch("os.path.getmtime")
    @patch("shutil.rmtree")
    @patch("os.path.exists")
    @patch.object(
        odcs.server.config.Config,
        "extra_target_dirs",
        new={"releng-private": "/tmp/private"},
    )
    def test_remove_old_koji_tag_cache_data(self, exists, rmtree, getmtime, listdir):
        exists.return_value = True
        now = time.time()
        listdir.side_effect = [["foo", "bar"], ["foo", "bar"]]
        # Default timeout is 30 days, so set 31 for first dir and 29 for
        # second dir.
        getmtime.side_effect = [now - 31 * 24 * 3600, now - 29 * 24 * 3600] * 2

        self.cache.remove_old_koji_tag_cache_data()
        self.assertEqual(
            rmtree.call_args_list,
            [
                call(os.path.join(self.cache.cache_dir, "foo")),
                call("/tmp/private/koji_tag_cache/foo"),
            ],
        )

    @patch("os.listdir")
    @patch("os.path.getmtime")
    @patch("shutil.rmtree")
    @patch("os.path.exists")
    def test_remove_old_koji_tag_cache_data_getmtime_raises(
        self, exists, rmtree, getmtime, listdir
    ):
        exists.return_value = True
        listdir.return_value = ["foo", "bar"]
        getmtime.side_effect = OSError("path does not exist")

        self.cache.remove_old_koji_tag_cache_data()
        rmtree.assert_not_called()



"""Add Compose.state_reason

Revision ID: a8e259e0208c
Revises: e2163db7b15d
Create Date: 2018-01-10 15:27:18.492001

"""

# revision identifiers, used by Alembic.
revision = "a8e259e0208c"
down_revision = "e2163db7b15d"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("state_reason", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "state_reason")



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import time
import os
import shutil
from datetime import timedelta
from flufl.lock import Lock

from odcs.common.types import PungiSourceType, COMPOSE_STATES
from odcs.server import conf, log
from odcs.server.utils import makedirs


class KojiTagCache(object):
    """
    Class caching old composes of KOJI_TAG source_type generated by Pungi.

    The goal of this caching is to let Pungi reuse old repodata when running
    createrepo_c and make composes faster.
    Cached composes are stored in the "conf.targetdir/koji_tag_cache" in
    directories named according to Koji tag from which the compose is
    generated, compose flags, sigkeys and arches. These are basically all the
    ODCS compose attributes which influences the repodata.
    """

    def __init__(self, compose):
        self.cache_dir = os.path.join(compose.target_dir, "koji_tag_cache")
        makedirs(self.cache_dir)

    @classmethod
    def remove_old_koji_tag_cache_data(klass):
        """
        Removes the old unused Koji tag cache directories which has not been
        used for more than `conf.koji_tag_cache_cleanup_timeout` days.
        """
        # The koji_tag_cache_cleanup_timeout is in days, so convert it to
        # seconds.
        older_than_seconds = conf.koji_tag_cache_cleanup_timeout * 24 * 3600
        threshold = time.time() - older_than_seconds

        for target_dir in [conf.target_dir] + list(conf.extra_target_dirs.values()):
            cache_dir = os.path.join(target_dir, "koji_tag_cache")
            if not os.path.exists(cache_dir):
                continue

            for cached_dir in os.listdir(cache_dir):
                path = os.path.join(cache_dir, cached_dir)
                try:
                    mtime = os.path.getmtime(path)
                except OSError:
                    # Directory might have been removed in meantime by some
                    # 3rd party process.
                    log.exception(
                        "Old koji tag cache directory %s removed while checking "
                        "Koji cache." % path
                    )
                    continue

                if mtime > threshold:
                    # Koji tag directory is not old enough to be removed.
                    continue

                log.info("Removing old Koji tag cache data in %s." % path)
                shutil.rmtree(path)

    def cached_compose_dir(self, compose):
        """
        Returns the full path to caching directory belonging to `compose`.
        The returned full path is based on Koji tag, flags, sigkeys and arches
        attributes of `compose`.

        :param models.Compose compose: Compose to generate the cache path from.
        :rtype: str
        :return: Full path to caching directory belonging to compose.
        """
        koji_tag = compose.source
        flags = str(compose.flags)
        if compose.sigkeys:
            sigkeys = "-".join(sorted(compose.sigkeys.split(" ")))
        else:
            sigkeys = ""
        if compose.arches:
            arches = "-".join(sorted(compose.arches.split(" ")))
        else:
            arches = ""
        dir_name = "-".join([koji_tag, flags, sigkeys, arches])
        return os.path.join(self.cache_dir, dir_name)

    def is_cached(self, compose):
        """
        Returns True if there is old compose generated from the same Koji tag
        which can be reused when generating this `compose`.

        :param models.Compose compose: Compose to check.
        :rtype: bool
        :return: Returns True when there exists cached compose for input
            compose.
        """
        if compose.source_type != PungiSourceType.KOJI_TAG:
            return False

        return os.path.exists(self.cached_compose_dir(compose))

    def reuse_cached(self, compose):
        """
        Configures the Cache directory so the Pungi will find the old compose
        defined by KojiTagCache.cached_compose_dir when building the `compose`.

        :param models.Compose compose: Compose which will reuse the older
            cached compose.
        """

        cached_compose_dir = self.cached_compose_dir(compose)
        log.info("Reusing repodata from old cached compose %s", cached_compose_dir)

        # Create the lock. The rmtree and copytree on same fs should not take more
        # than 3 minutes really.
        lock = Lock(cached_compose_dir + ".lock")
        lock.lifetime = timedelta(minutes=3)

        try:
            lock.lock()

            compose_dir_name = "odcs-%d-1-19700101.n.0" % compose.id
            compose_dir = os.path.join(self.cache_dir, compose_dir_name)

            if os.path.exists(compose_dir):
                shutil.rmtree(compose_dir)
            shutil.copytree(cached_compose_dir, compose_dir, symlinks=True)
        finally:
            if lock.is_locked:
                lock.unlock()

    def cleanup_reused(self, compose):
        """
        Cleans-up the configuration done by `reuse_cached`. Should be called
        once the image depending on cached image is built.
        """
        compose_dir_name = "odcs-%d-1-19700101.n.0" % compose.id
        compose_dir = os.path.join(self.cache_dir, compose_dir_name)

        if os.path.exists(compose_dir):
            log.info("Removing cached reused compose %s" % compose_dir)
            shutil.rmtree(compose_dir)

    def update_cache(self, compose):
        """
        Updates the cache with newly built version of `compose`.

        :param models.Compose compose: Compose to update the cache from.
        """
        if (
            compose.source_type != PungiSourceType.KOJI_TAG
            or compose.state != COMPOSE_STATES["done"]
        ):
            log.info("Not caching the compose %s.", compose)
            return

        log.info("Caching the compose %s", compose)
        cached_compose_dir = self.cached_compose_dir(compose)
        compose_dir = os.path.realpath(compose.toplevel_dir)

        # Create the lock. The rmtree and copytree on same fs should not take more
        # than 3 minutes really.
        lock = Lock(cached_compose_dir + ".lock")
        lock.lifetime = timedelta(minutes=3)

        try:
            lock.lock()
            if os.path.exists(cached_compose_dir):
                shutil.rmtree(cached_compose_dir)
            shutil.copytree(compose_dir, cached_compose_dir, symlinks=True)
        finally:
            if lock.is_locked:
                lock.unlock()



"""Add arches field to composes

Revision ID: 11b350234051
Revises: a8e259e0208c
Create Date: 2018-02-14 07:33:29.739034

"""

# revision identifiers, used by Alembic.
revision = "11b350234051"
down_revision = "a8e259e0208c"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("arches", sa.String(), nullable=True, default="x86_64")
    )


def downgrade():
    op.drop_column("composes", "arches")



"""Add Compose.builds.

Revision ID: 4514febd31fa
Revises: d1da07e15c54
Create Date: 2018-10-25 13:28:01.798873

"""

# revision identifiers, used by Alembic.
revision = "4514febd31fa"
down_revision = "d1da07e15c54"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("builds", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "builds")



"""Add koji_task_id

Revision ID: e2163db7b15d
Revises: b75ad2afc207
Create Date: 2017-12-13 08:42:05.900658

"""

# revision identifiers, used by Alembic.
revision = "e2163db7b15d"
down_revision = "b75ad2afc207"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("koji_task_id", sa.Integer(), nullable=True))
    op.create_index(
        op.f("ix_composes_koji_task_id"), "composes", ["koji_task_id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_composes_koji_task_id"), table_name="composes")
    op.drop_column("composes", "koji_task_id")



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import datetime

from flask.views import MethodView, View
from flask import render_template, request, jsonify, g, Response
from werkzeug.exceptions import BadRequest
from prometheus_client import generate_latest, CONTENT_TYPE_LATEST

from odcs.server import app, db, log, conf, version
from odcs.server.errors import NotFound, Forbidden
from odcs.server.models import Compose
from odcs.common.types import (
    COMPOSE_RESULTS,
    COMPOSE_FLAGS,
    COMPOSE_STATES,
    PUNGI_SOURCE_TYPE_NAMES,
    PungiSourceType,
    MULTILIB_METHODS,
)
from odcs.server.api_utils import (
    pagination_metadata,
    filter_composes,
    validate_json_data,
    raise_if_input_not_allowed,
    cors_header,
)
from odcs.server.auth import requires_role, login_required, has_role
from odcs.server.auth import require_scopes
from odcs.server.metrics import registry

try:
    from odcs.server.celery_tasks import celery_app, schedule_compose

    CELERY_AVAILABLE = True
except ImportError:
    log.exception("Cannot import celery_tasks. The Celery support is turned off.")
    CELERY_AVAILABLE = False


api_v1 = {
    "composes": {
        "url": "/api/1/composes/",
        "options": {"defaults": {"id": None}, "methods": ["GET"]},
    },
    "compose": {"url": "/api/1/composes/<int:id>", "options": {"methods": ["GET"]}},
    "composes_post": {"url": "/api/1/composes/", "options": {"methods": ["POST"]}},
    "compose_regenerate": {
        "url": "/api/1/composes/<int:id>",
        "options": {"methods": ["PATCH"]},
    },
    "composes_delete": {
        "url": "/api/1/composes/<int:id>",
        "options": {"methods": ["DELETE"]},
    },
    "about": {"url": "/api/1/about/", "options": {"methods": ["GET"]}},
    "metrics": {"url": "/api/1/metrics/", "options": {"methods": ["GET"]}},
}


class ODCSAPI(MethodView):
    def _get_compose_owner(self):
        if conf.auth_backend == "noauth":
            log.warning(
                "Cannot determine the owner of compose, because "
                "'noauth' auth_backend is used."
            )
            return "unknown"
        else:
            return g.user.username

    def _get_seconds_to_live(self, request_data):
        seconds_to_live = request_data.get("seconds_to_live")
        # Fallback to old name of this variable to keep backward compatibility.
        if seconds_to_live is None:
            seconds_to_live = request_data.get("seconds-to-live")
        if seconds_to_live:
            try:
                return min(int(seconds_to_live), conf.max_seconds_to_live)
            except ValueError:
                err = "Invalid seconds_to_live specified in request: %s" % request_data
                log.error(err)
                raise ValueError(err)
        else:
            return conf.seconds_to_live

    @cors_header()
    def get(self, id):
        """Returns ODCS composes.

        If ``id`` is set, only the ODCS compose defined by that ID is
        returned.

        :query string owner: Return only composes owned by this:ref:`owner<owner>`.
        :query number source_type: Return only composes of this :ref:`source type<source_type>`.
        :query string source: Return only composes built from this :ref:`source<source>`.
        :query number state: Return only composes built from this :ref:`state<state>`.
        :query string order_by: Order the composes by the given field. If ``-`` prefix is used,
            the order will be descending. The default value is ``-id``. Available fields are:

            - :ref:`id<id>`
            - :ref:`owner<owner>`
            - :ref:`source_type<source_type>`
            - :ref:`koji_event<koji_event>`
            - :ref:`state<state>`
            - :ref:`time_to_expire<time_to_expire>`
            - :ref:`time_done<time_done>`
            - :ref:`time_removed<time_removed>`

        :statuscode 200: Composes are returned.
        :statuscode 404: Compose not found.
        """
        if id is None:
            p_query = filter_composes(request)

            json_data = {
                "meta": pagination_metadata(p_query, request.args),
                "items": [item.json() for item in p_query.items],
            }

            return jsonify(json_data), 200

        else:
            compose = Compose.query.filter_by(id=id).first()
            if compose:
                return jsonify(compose.json(True)), 200
            else:
                raise NotFound("No such compose found.")

    @login_required
    @require_scopes("renew-compose")
    @requires_role("allowed_clients")
    def patch(self, id):
        """Extends the compose expiration time or regenerates expired compose.

        :query number id: :ref:`ID<id>` of the compose to update/regenerate.
        :jsonparam number seconds_to_live: Number of seconds before the compoose expires.
        :jsonparam list sigkeys: Optional list defining the :ref:`sigkeys<sigkeys>`.
            If not defined, original :ref:`sigkeys<sigkeys>` will be used.

        :statuscode 200: Compose updated and returned.
        :statuscode 401: User is unathorized.
        :statuscode 404: Compose not found.
        """
        if request.data:
            data = request.get_json(force=True)
        else:
            data = {}
        validate_json_data(data)

        seconds_to_live = self._get_seconds_to_live(data)

        old_compose = Compose.query.filter(
            Compose.id == id,
            Compose.state.in_(
                [
                    COMPOSE_STATES["removed"],
                    COMPOSE_STATES["done"],
                    COMPOSE_STATES["failed"],
                ]
            ),
        ).first()

        if not old_compose:
            err = "No compose with id %s found" % id
            log.error(err)
            raise NotFound(err)

        # Backward compatibility for old composes which don't have
        # the compose_type set - we treat them as "test" composes
        # when regenerating them.
        compose_type = old_compose.compose_type or "test"

        sigkeys = ""
        if "sigkeys" in data:
            sigkeys = " ".join(data["sigkeys"])
        else:
            sigkeys = old_compose.sigkeys

        # Get the raw_config_key if this compose is RAW_CONFIG.
        raw_config_key = None
        if old_compose.source_type == PungiSourceType.RAW_CONFIG:
            raw_config_key = old_compose.source.split("#")[0]

        raise_if_input_not_allowed(
            source_types=old_compose.source_type,
            sources=old_compose.source,
            results=old_compose.results,
            flags=old_compose.flags,
            arches=old_compose.arches,
            compose_types=compose_type,
            raw_config_keys=raw_config_key,
        )

        has_to_create_a_copy = (
            old_compose.state in (COMPOSE_STATES["removed"], COMPOSE_STATES["failed"])
            or sigkeys != old_compose.sigkeys
        )
        if has_to_create_a_copy:
            log.info("%r: Going to regenerate the compose", old_compose)
            compose = Compose.create_copy(
                db.session,
                old_compose,
                self._get_compose_owner(),
                seconds_to_live,
                sigkeys=sigkeys,
            )
            db.session.add(compose)
            # Flush is needed, because we use `before_commit` SQLAlchemy
            # event to send message and before_commit can be called before
            # flush and therefore the compose ID won't be set.
            db.session.flush()
            db.session.commit()

            if CELERY_AVAILABLE and conf.celery_broker_url:
                schedule_compose(compose)

            return jsonify(compose.json()), 200
        else:
            # Otherwise, just extend expiration to make it usable for longer
            # time.
            extend_from = datetime.datetime.utcnow()
            old_compose.extend_expiration(extend_from, seconds_to_live)
            log.info(
                "Extended time_to_expire for compose %r to %s",
                old_compose,
                old_compose.time_to_expire,
            )
            # As well as extending those composes that reuse this this compose,
            # and the one this compose reuses.
            reused_compose = old_compose.get_reused_compose()
            if reused_compose:
                reused_compose.extend_expiration(extend_from, seconds_to_live)
            for c in old_compose.get_reusing_composes():
                c.extend_expiration(extend_from, seconds_to_live)
            db.session.commit()
            return jsonify(old_compose.json()), 200

    @login_required
    @require_scopes("new-compose")
    @requires_role("allowed_clients")
    def post(self):
        """Creates new ODCS compose request.

        :jsonparam number seconds_to_live: Number of seconds before the compoose expires.
        :jsonparam list flags: List of :ref:`compose flags<flags>` defined as strings.
        :jsonparam list arches: List of :ref:`arches<arches>` the compose should be generated for.
        :jsonparam list multilib_arches: List of :ref:`multilib arches<multilib_arches>`.
        :jsonparam string multilib_method: List defining the :ref:`multilib method<multilib_method>`.
        :jsonparam list lookaside_repos: List of :ref:`lookaside_repos<lookaside_repos>`.
        :jsonparam string label: String defining the :ref:`label<label>`.
        :jsonparam string compose_type: String defining the :ref:`compose_type<compose_type>`.
        :jsonparam string target_dir: String defining the :ref:`target_dir<target_dir>`.
        :jsonparam list parent_pungi_compose_ids: Pungi compose IDs of parent composes associated
            with this compose. They will be stored with this compose in the Compose Tracking
            Service.
        :jsonparam object source: The JSON object defining the source of compose.
        :jsonparam string source["type"]: String defining the :ref:`source type<source_type>`.
        :jsonparam string source["source"]: String defining the :ref:`source<source>`.
        :jsonparam list source["packages"]: List defining the :ref:`packages<packages>`.
        :jsonparam list source["builds"]: List defining the :ref:`builds<builds>`.
        :jsonparam list source["sigkeys"]: List defining the :ref:`sigkeys<sigkeys>`.
        :jsonparam list source["koji_event"]: Number defining the :ref:`koji_event<koji_event>`.
        :jsonparam list source["modular_koji_tags"]: List defining the :ref:`modular_koji_tags<modular_koji_tags>`.
        :jsonparam list source["scratch_modules"]: List defining the :ref:`scratch_modules<scratch_modules>`.
        :jsonparam list source["scratch_build_tasks"]: List defining the :ref:`scratch_build_tasks<scratch_build_tasks>`.

        :statuscode 200: Compose request created and returned.
        :statuscode 401: Request not in valid format.
        :statuscode 401: User is unathorized.
        """
        data = request.get_json(force=True)
        if not data:
            raise ValueError("No JSON POST data submitted")

        validate_json_data(data)

        seconds_to_live = self._get_seconds_to_live(data)

        source_data = data.get("source", None)
        if not isinstance(source_data, dict):
            err = "Invalid source configuration provided: %s" % str(data)
            log.error(err)
            raise ValueError(err)

        needed_keys = ["type"]
        for key in needed_keys:
            if key not in source_data:
                err = "Missing %s in source configuration, received: %s" % (
                    key,
                    str(source_data),
                )
                log.error(err)
                raise ValueError(err)

        source_type = source_data["type"]
        if source_type not in PUNGI_SOURCE_TYPE_NAMES:
            err = 'Unknown source type "%s"' % source_type
            log.error(err)
            raise ValueError(err)

        source_type = PUNGI_SOURCE_TYPE_NAMES[source_type]

        source = []
        if "source" in source_data and source_data["source"] != "":
            # Use list(set()) here to remove duplicate sources.
            source = list(set(source_data["source"].split(" ")))

        scratch_modules = None
        if "scratch_modules" in source_data:
            scratch_modules = " ".join(source_data["scratch_modules"])

        if (
            not source
            and source_type != PungiSourceType.BUILD
            and not (source_type == PungiSourceType.MODULE and scratch_modules)
        ):
            err = "No source provided for %s" % source_type
            log.error(err)
            raise ValueError(err)

        # Validate `source` based on `source_type`.
        raw_config_key = None
        if source_type == PungiSourceType.RAW_CONFIG:
            if len(source) > 1:
                raise ValueError(
                    'Only single source is allowed for "raw_config" ' "source_type"
                )

            source_name_hash = source[0].split("#")
            if (
                len(source_name_hash) != 2
                or not source_name_hash[0]
                or not source_name_hash[1]
            ):
                raise ValueError(
                    'Source must be in "source_name#commit_hash" format for '
                    '"raw_config" source_type.'
                )

            source_name, source_hash = source_name_hash
            if source_name not in conf.raw_config_urls:
                raise ValueError(
                    'Source "%s" does not exist in server configuration.' % source_name
                )
            raw_config_key = source_name
        elif source_type == PungiSourceType.MODULE:
            for module_str in source:
                nsvc = module_str.split(":")
                if len(nsvc) < 2:
                    raise ValueError(
                        'Module definition must be in "n:s", "n:s:v" or '
                        '"n:s:v:c" format, but got %s' % module_str
                    )
                if nsvc[0] in conf.base_module_names:
                    raise ValueError(
                        "ODCS currently cannot create compose with base "
                        "modules, but %s was requested." % nsvc[0]
                    )

        source = " ".join(source)

        packages = None
        if "packages" in source_data:
            packages = " ".join(source_data["packages"])

        builds = None
        if "builds" in source_data:
            builds = " ".join(source_data["builds"])

        sigkeys = ""
        if "sigkeys" in source_data:
            sigkeys = " ".join(source_data["sigkeys"])
        else:
            sigkeys = " ".join(conf.sigkeys)

        koji_event = source_data.get("koji_event", None)

        flags = 0
        if "flags" in data:
            for name in data["flags"]:
                if name not in COMPOSE_FLAGS:
                    raise ValueError("Unknown flag %s", name)
                flags |= COMPOSE_FLAGS[name]

        results = COMPOSE_RESULTS["repository"]
        if "results" in data:
            for name in data["results"]:
                if name not in COMPOSE_RESULTS:
                    raise ValueError("Unknown result %s", name)
                results |= COMPOSE_RESULTS[name]

        arches = None
        if "arches" in data:
            arches = " ".join(data["arches"])
        else:
            arches = " ".join(conf.arches)

        multilib_arches = ""
        if "multilib_arches" in data:
            multilib_arches = " ".join(data["multilib_arches"])

        lookaside_repos = ""
        if "lookaside_repos" in data:
            lookaside_repos = " ".join(data["lookaside_repos"])

        parent_pungi_compose_ids = None
        if "parent_pungi_compose_ids" in data:
            parent_pungi_compose_ids = " ".join(data["parent_pungi_compose_ids"])

        multilib_method = MULTILIB_METHODS["none"]
        if "multilib_method" in data:
            for name in data["multilib_method"]:
                if name not in MULTILIB_METHODS:
                    raise ValueError('Unknown multilib method "%s"' % name)
                multilib_method |= MULTILIB_METHODS[name]

        modular_koji_tags = None
        if "modular_koji_tags" in source_data:
            modular_koji_tags = " ".join(source_data["modular_koji_tags"])

        module_defaults_url = None
        if "module_defaults_url" in source_data:
            module_defaults_url = source_data["module_defaults_url"]

        module_defaults_commit = None
        if "module_defaults_commit" in source_data:
            module_defaults_commit = source_data["module_defaults_commit"]

        module_defaults = None
        # The "^" operator is logical XOR.
        if bool(module_defaults_url) ^ bool(module_defaults_commit):
            raise ValueError(
                'The "module_defaults_url" and "module_defaults_commit" '
                "must be used together."
            )
        elif module_defaults_url and module_defaults_commit:
            module_defaults = "%s %s" % (module_defaults_url, module_defaults_commit)

        scratch_build_tasks = None
        if "scratch_build_tasks" in source_data:
            scratch_build_tasks = " ".join(source_data["scratch_build_tasks"])

        label = data.get("label", None)
        compose_type = data.get("compose_type", "test")

        target_dir = data.get("target_dir")
        if target_dir and target_dir != "default":
            if target_dir not in conf.extra_target_dirs:
                raise ValueError('Unknown "target_dir" "%s"' % target_dir)
            target_dir = conf.extra_target_dirs[target_dir]
        else:
            target_dir = conf.target_dir

        raise_if_input_not_allowed(
            source_types=source_type,
            sources=source,
            results=results,
            flags=flags,
            arches=arches,
            compose_types=compose_type,
            target_dirs=target_dir,
            raw_config_keys=raw_config_key,
        )

        compose = Compose.create(
            db.session,
            self._get_compose_owner(),
            source_type,
            source,
            results,
            seconds_to_live,
            packages,
            flags,
            sigkeys,
            koji_event,
            arches,
            multilib_arches=multilib_arches,
            multilib_method=multilib_method,
            builds=builds,
            lookaside_repos=lookaside_repos,
            modular_koji_tags=modular_koji_tags,
            module_defaults_url=module_defaults,
            label=label,
            compose_type=compose_type,
            target_dir=target_dir,
            scratch_modules=scratch_modules,
            parent_pungi_compose_ids=parent_pungi_compose_ids,
            scratch_build_tasks=scratch_build_tasks,
        )
        db.session.add(compose)
        # Flush is needed, because we use `before_commit` SQLAlchemy event to
        # send message and before_commit can be called before flush and
        # therefore the compose ID won't be set.
        db.session.flush()
        db.session.commit()

        if CELERY_AVAILABLE and conf.celery_broker_url:
            schedule_compose(compose)

        return jsonify(compose.json()), 200

    @login_required
    @require_scopes("delete-compose")
    def delete(self, id):
        """Cancels waiting compose or marks finished compose as expired to be
        removed later from ODCS storage. The compose metadata are still stored
        in the ODCS database, only the composed files stored in ODCS storage
        are removed.

        Users are allowed to cancel their own composes. Deleting is limited to
        admins. Admins can also cancel any compose.

        :query number id: :ref:`ID<id>` of the compose to delete.

        :statuscode 202: Compose updated and returned.
        :statuscode 400: Compose is not in wait, done or failed :ref:`state<state>`.
        :statuscode 401: User is unathorized.
        :statuscode 402: User doesn't own the compose to be cancelled or is not admin
        :statuscode 404: Compose not found.
        """
        compose = Compose.query.filter_by(id=id).first()
        if not compose:
            raise NotFound("No such compose found.")

        is_admin = has_role("admins")

        # First try to cancel the compose
        if CELERY_AVAILABLE and compose.state == COMPOSE_STATES["wait"]:
            if not is_admin and compose.owner != g.user.username:
                raise Forbidden(
                    "Compose (id=%s) can not be canceled, it's owned by someone else."
                )

            # Revoke the task
            if compose.celery_task_id:
                celery_app.control.revoke(compose.celery_task_id)
            # Change compose status to failed
            compose.transition(
                COMPOSE_STATES["failed"], "Canceled by %s" % g.user.username
            )
            message = "Compose (id=%s) has been canceled" % id
            response = jsonify({"status": 202, "message": message})
            response.status_code = 202
            return response

        # Compose was not eligible for cancellation, try to delete it instead.
        # Only admins can do this.
        if not is_admin:
            raise Forbidden("User %s is not in role admins." % g.user.username)

        # can remove compose that is in state of 'done' or 'failed'
        deletable_states = {n: COMPOSE_STATES[n] for n in ["done", "failed"]}
        if compose.state not in deletable_states.values():
            raise BadRequest(
                "Compose (id=%s) can not be removed, its state need to be in %s."
                % (id, deletable_states.keys())
            )

        # change compose.time_to_expire to now, so backend will
        # delete this compose as it's an expired compose now
        compose.time_to_expire = datetime.datetime.utcnow()
        compose.removed_by = g.user.username
        db.session.add(compose)
        db.session.commit()
        message = (
            "The delete request for compose (id=%s) has been accepted and will be"
            " processed by backend later." % compose.id
        )
        response = jsonify({"status": 202, "message": message})
        response.status_code = 202
        return response


class AboutAPI(MethodView):
    @cors_header()
    def get(self):
        """Returns information about this ODCS instance in JSON format.

        :resjson string version: The ODCS server version.
        :resjson string auth_backend: The name of authorization backend this
            server is configured with. Can be one of following:

            - ``noauth`` - No authorization is required.
            - ``kerberos`` - Kerberos authorization is required.
            - ``openidc`` - OpenIDC authorization is required.
            - ``kerberos_or_ssl`` - Kerberos or SSL authorization is required.
            - ``ssl`` - SSL authorization is required.
        :resjson dict allowed_clients: Allowed clients in the same format as
            in the ODCS server configuration file.
        :resjson dict raw_config_urls: Raw config URLs in the same format as
            in the ODCS server configuration file.
        :resjson list sigkeys: Default list of sigkeys.
        :statuscode 200: Compose updated and returned.
        """
        json = {"version": version}
        config_items = ["auth_backend", "allowed_clients", "raw_config_urls", "sigkeys"]
        for item in config_items:
            config_item = getattr(conf, item)
            # All config items have a default, so if doesn't exist it is
            # an error
            if config_item is None:
                raise ValueError('An invalid config item of "%s" was specified' % item)
            json[item] = config_item
        return jsonify(json), 200


class Index(View):

    methods = ["GET"]

    def dispatch_request(self):
        return render_template("index.html")


class MetricsAPI(MethodView):
    @cors_header()
    def get(self):
        """
        Returns the Prometheus metrics.

        :statuscode 200: Prometheus metrics returned.
        """
        return Response(generate_latest(registry), content_type=CONTENT_TYPE_LATEST)


def register_api_v1():
    """ Registers version 1 of ODCS API. """
    composes_view = ODCSAPI.as_view("composes")
    about_view = AboutAPI.as_view("about")
    metrics_view = MetricsAPI.as_view("metrics")
    for key, val in api_v1.items():
        if key.startswith("compose"):
            app.add_url_rule(
                val["url"], endpoint=key, view_func=composes_view, **val["options"]
            )
        elif key.startswith("about"):
            app.add_url_rule(
                val["url"], endpoint=key, view_func=about_view, **val["options"]
            )
        elif key.startswith("metrics"):
            app.add_url_rule(
                val["url"], endpoint=key, view_func=metrics_view, **val["options"]
            )
        else:
            raise ValueError("Unhandled API key: %s." % key)


app.add_url_rule("/", view_func=Index.as_view("index"))
register_api_v1()



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import unittest
import copy

import mock
from mock import patch, Mock, MagicMock

from odcs.client.odcs import AuthMech
from odcs.client.odcs import (
    ODCS,
    ComposeSourceTag,
    ComposeSourceModule,
    ComposeSourcePulp,
    ComposeSourceRawConfig,
    ComposeSourceBuild,
    ComposeLog,
)
from odcs.client.odcs import validate_int


class TestValidateIntValue(unittest.TestCase):
    """Test validate_int"""

    def test_failure_validate_int(self):
        self.assertRaises(TypeError, validate_int, "a")
        self.assertRaises(TypeError, validate_int, "1")
        self.assertRaises(TypeError, validate_int, "")
        self.assertRaises(TypeError, validate_int, None)
        self.assertRaises(ValueError, validate_int, 0)
        self.assertRaises(ValueError, validate_int, -1)

    def test_succeed_validate_int(self):
        self.assertEqual(None, validate_int(100))
        self.assertEqual(None, validate_int(1))


class TestMakeEndPoint(unittest.TestCase):
    """Test ODCS._make_endpoint"""

    def setUp(self):
        self.server_url = "http://localhost/"

    def test_make_endpoint(self):
        odcs = ODCS(self.server_url)

        endpoint_url = odcs._make_endpoint("composes/")
        self.assertEqual(
            "{0}api/{1}/composes/".format(self.server_url, odcs.api_version),
            endpoint_url,
        )

        endpoint_url = odcs._make_endpoint("/composes/")
        self.assertEqual(
            "{0}api/{1}/composes/".format(self.server_url, odcs.api_version),
            endpoint_url,
        )


class TestMakeRequest(unittest.TestCase):
    """Test ODCS._make_request"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.resource_path = "composes/"

    @patch("odcs.client.odcs.requests")
    def test_raise_error(self, requests):
        requests.get.return_value.status_code = 401

        self.odcs = ODCS(self.server_url)
        r = self.odcs._make_request("get", self.resource_path)

        self.assertEqual(requests.get.return_value, r)
        requests.get.return_value.raise_for_status.assert_called_once()

    @patch("odcs.client.odcs.requests")
    @patch("odcs.client.odcs.HTTPKerberosAuth")
    def test_with_kerberos_auth(self, HTTPKerberosAuth, requests):
        requests.get.return_value.status_code = 200
        expected_auth = HTTPKerberosAuth.return_value

        odcs = ODCS(self.server_url, auth_mech=AuthMech.Kerberos)
        r = odcs._make_request("get", self.resource_path)

        self.assertEqual(requests.get.return_value, r)
        requests.get.assert_called_once_with(
            odcs._make_endpoint(self.resource_path), auth=expected_auth
        )

    @patch("odcs.client.odcs.requests")
    def test_with_openidc_auth(self, requests):
        fake_openidc_token = "1234567890"
        requests.post.return_value.status_code = 200

        odcs = ODCS(
            self.server_url,
            auth_mech=AuthMech.OpenIDC,
            openidc_token=fake_openidc_token,
        )
        r = odcs._make_request("post", self.resource_path, data={"id": 1})

        self.assertEqual(requests.post.return_value, r)
        requests.post.assert_called_once_with(
            odcs._make_endpoint(self.resource_path),
            data=json.dumps({"id": 1}),
            headers={
                "Authorization": "Bearer {0}".format(fake_openidc_token),
                "Content-Type": "application/json",
            },
        )

    @patch("odcs.client.odcs.requests")
    def test_with_ssl_auth(self, requests):
        requests.post.return_value.status_code = 200

        odcs = ODCS(
            self.server_url,
            auth_mech=AuthMech.SSL,
            ssl_cert="./ssl.crt",
            ssl_key="./ssl.key",
        )
        r = odcs._make_request("post", self.resource_path, data={"id": 1})

        self.assertEqual(requests.post.return_value, r)
        requests.post.assert_called_once_with(
            odcs._make_endpoint(self.resource_path),
            data=json.dumps({"id": 1}),
            cert=("./ssl.crt", "./ssl.key"),
            headers={"Content-Type": "application/json"},
        )

    @patch("odcs.client.odcs.requests")
    def test_do_not_verify_ssl(self, requests):
        requests.post.return_value.status_code = 200

        odcs = ODCS(self.server_url, verify_ssl=False)
        r = odcs._make_request("post", self.resource_path)

        self.assertEqual(requests.post.return_value, r)
        requests.post.assert_called_once_with(
            odcs._make_endpoint(self.resource_path), verify=False
        )


class TestGetCompose(unittest.TestCase):
    """Test ODCS.get_compose"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    @patch("odcs.client.odcs.requests")
    def test_get_compose(self, requests):
        fake_compose = {
            "flags": [],
            "id": 1,
            "owner": "Unknown",
            "result_repo": "http://odcs.host.qe.eng.pek2.redhat.com/composes/latest-odcs-1-1/compose/Temporary",
            "source": "cf-1.0-rhel-5",
            "source_type": 1,
            "state": 4,
            "state_name": "failed",
            "time_done": "2017-07-11T13:05:40Z",
            "time_removed": None,
            "time_submitted": "2017-07-11T13:05:40Z",
            "time_to_expire": "2017-07-12T13:05:40Z",
        }
        requests.get = Mock()
        requests.get.return_value.status_code = 200
        requests.get.return_value.json.return_value = fake_compose

        compose = self.odcs.get_compose(1)

        self.assertEqual(fake_compose, compose)
        requests.get.assert_called_once_with(
            "{0}api/{1}/composes/1".format(self.server_url, self.odcs.api_version)
        )


class TestDeleteCompose(unittest.TestCase):
    """Test ODCS.delete_compose"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    @patch("odcs.client.odcs.requests")
    def test_compose_id_not_found(self, requests):
        fake_response = {
            "status": 404,
            "error": "Not Found",
            "message": "No such compose found.",
        }
        requests.delete.return_value.status_code = 404
        requests.delete.return_value.json.return_value = fake_response

        self.odcs.delete_compose(1)

        requests.delete.return_value.raise_for_status.assert_called_once()

    @patch("odcs.client.odcs.requests")
    def test_delete_compose(self, requests):
        fake_response = {
            "status": 202,
            "message": "The delete request for compose (id=1) has been accepted",
        }
        requests.delete.return_value.status_code = 202
        requests.delete.return_value.json.return_value = fake_response

        r = self.odcs.delete_compose(1)

        self.assertEqual(fake_response, r)
        requests.delete.assert_called_once_with(self.odcs._make_endpoint("composes/1"))


@patch("odcs.client.odcs.requests")
class TestNewCompose(unittest.TestCase):
    """Test ODCS.new_compose"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    def test_create_a_new_compose(self, requests):
        fake_new_compose = {
            "flags": [],
            "id": 7,
            "owner": "Unknown",
            "result_repo": "http://odcs.host.qe.eng.pek2.redhat.com/composes/latest-odcs-7-1/compose/Temporary",
            "source": "cf-1.0-rhel-5",
            "source_type": 1,
            "koji_event": 123456,
            "state": 0,
            "state_name": "wait",
            "time_done": None,
            "time_removed": None,
            "time_submitted": "2017-07-21T03:33:43Z",
            "time_to_expire": "2017-07-22T03:33:43Z",
        }
        requests.post.return_value.status_code = 200
        requests.post.return_value.json.return_value = fake_new_compose

        new_compose = self.odcs.new_compose(
            "cf-1.0-rhel-5",
            "tag",
            packages=["libdnet"],
            sigkeys=["123", "456"],
            koji_event=123456,
            results=["boot.iso"],
        )

        self.assertEqual(fake_new_compose, new_compose)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps(
                {
                    "source": {
                        "source": "cf-1.0-rhel-5",
                        "type": "tag",
                        "packages": ["libdnet"],
                        "sigkeys": ["123", "456"],
                        "koji_event": 123456,
                    },
                    "results": ["boot.iso"],
                }
            ),
            headers={"Content-Type": "application/json"},
        )

    def test_request_compose_source_tag(self, requests):
        source = ComposeSourceTag("f32")
        self.odcs.request_compose(source)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps({"source": {"source": "f32", "type": "tag"}}),
            headers={"Content-Type": "application/json"},
        )

    def test_request_compose_source_module(self, requests):
        source = ComposeSourceModule(["testmodule:master", "foo:bar"])
        self.odcs.request_compose(source)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps(
                {"source": {"source": "testmodule:master foo:bar", "type": "module"}}
            ),
            headers={"Content-Type": "application/json"},
        )

    def test_request_compose_source_pulp(self, requests):
        source = ComposeSourcePulp(["content-set1", "content-set2"])
        self.odcs.request_compose(source)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps(
                {"source": {"source": "content-set1 content-set2", "type": "pulp"}}
            ),
            headers={"Content-Type": "application/json"},
        )

    def test_request_compose_source_raw_config(self, requests):
        source = ComposeSourceRawConfig("name", "commit")
        self.odcs.request_compose(source)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps(
                {"source": {"source": "name#commit", "type": "raw_config"}}
            ),
            headers={"Content-Type": "application/json"},
        )

    def test_request_compose_source_build(self, requests):
        source = ComposeSourceBuild(["foo-1-1", "bar-1-1"])
        self.odcs.request_compose(source)
        requests.post.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            data=json.dumps(
                {
                    "source": {
                        "source": "",
                        "type": "build",
                        "builds": ["foo-1-1", "bar-1-1"],
                    },
                }
            ),
            headers={"Content-Type": "application/json"},
        )


class TestRenewCompose(unittest.TestCase):
    """Test ODCS.renew_compose"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    @patch("odcs.client.odcs.requests")
    def test_renew_a_compose(self, requests):
        fake_renew_compose = {
            "flags": [],
            "id": 7,
            "owner": "Unknown",
            "result_repo": "http://odcs.host.qe.eng.pek2.redhat.com/composes/latest-odcs-7-1/compose/Temporary",
            "source": "cf-1.0-rhel-5",
            "source_type": 1,
            "state": 0,
            "state_name": "wait",
            "time_done": None,
            "time_removed": None,
            "time_submitted": "2017-07-21T03:33:43Z",
            "time_to_expire": "2017-07-22T03:33:43Z",
        }
        requests.patch.return_value.status_code = 200
        requests.patch.return_value.json.return_value = fake_renew_compose

        r = self.odcs.renew_compose(6, seconds_to_live=60)

        self.assertEqual(fake_renew_compose, r)
        requests.patch.assert_called_once_with(
            self.odcs._make_endpoint("composes/6"),
            data=json.dumps({"seconds-to-live": 60}),
            headers={"Content-Type": "application/json"},
        )


class TestFindComposes(unittest.TestCase):
    """Test ODCS.find_composes"""

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    @patch("odcs.client.odcs.requests")
    def test_find_composes_without_pagination(self, requests):
        fake_found_composes = {
            "items": [
                {
                    "flags": [],
                    "id": 1,
                    "owner": "Unknown",
                    "result_repo": "http://localhost/composes/latest-odcs-1-1/compose/Temporary",
                    "source": "cf-1.0-rhel-5",
                    "source_type": 1,
                    "state": 4,
                    "state_name": "failed",
                    "time_done": "2017-07-11T13:05:40Z",
                    "time_removed": None,
                    "time_submitted": "2017-07-11T13:05:40Z",
                    "time_to_expire": "2017-07-12T13:05:40Z",
                },
                {
                    "flags": [],
                    "id": 2,
                    "owner": "Unknown",
                    "result_repo": "http://localhost/composes/latest-odcs-2-1/compose/Temporary",
                    "source": "cf-1.0-rhel-5",
                    "source_type": 1,
                    "state": 4,
                    "state_name": "failed",
                    "time_done": "2017-07-11T13:07:42Z",
                    "time_removed": None,
                    "time_submitted": "2017-07-11T13:07:41Z",
                    "time_to_expire": "2017-07-12T13:07:41Z",
                },
            ],
            "meta": {"page": 1, "pages": 1},
        }
        requests.get.return_value.status_code = 200
        requests.get.return_value.json.return_value = fake_found_composes

        r = self.odcs.find_composes(owner="unknown", source_type="tag")

        self.assertEqual(fake_found_composes, r)
        requests.get.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            params={"owner": "unknown", "source_type": "tag"},
        )

    @patch("odcs.client.odcs.requests")
    def test_find_composes_the_second_page(self, requests):
        fake_found_composes = {
            "items": [
                {
                    "flags": [],
                    "id": 1,
                    "owner": "Unknown",
                    "result_repo": "http://localhost/composes/latest-odcs-1-1/compose/Temporary",
                    "source": "cf-1.0-rhel-5",
                    "source_type": 1,
                    "state": 4,
                    "state_name": "failed",
                    "time_done": "2017-07-11T13:05:40Z",
                    "time_removed": None,
                    "time_submitted": "2017-07-11T13:05:40Z",
                    "time_to_expire": "2017-07-12T13:05:40Z",
                },
                {
                    "flags": [],
                    "id": 2,
                    "owner": "Unknown",
                    "result_repo": "http://localhost/composes/latest-odcs-2-1/compose/Temporary",
                    "source": "cf-1.0-rhel-5",
                    "source_type": 1,
                    "state": 4,
                    "state_name": "failed",
                    "time_done": "2017-07-11T13:07:42Z",
                    "time_removed": None,
                    "time_submitted": "2017-07-11T13:07:41Z",
                    "time_to_expire": "2017-07-12T13:07:41Z",
                },
            ],
            "meta": {"page": 1, "pages": 1},
        }
        requests.get.return_value.status_code = 200
        requests.get.return_value.json.return_value = fake_found_composes

        r = self.odcs.find_composes(owner="unknown", source_type="tag", page=2)

        self.assertEqual(fake_found_composes, r)
        requests.get.assert_called_once_with(
            self.odcs._make_endpoint("composes/"),
            params={"owner": "unknown", "source_type": "tag", "page": 2},
        )


@patch("time.sleep")
@patch("odcs.client.odcs.ODCS.get_compose")
class TestWaitForCompose(unittest.TestCase):
    """Test ODCS.wait_for_compose"""

    _TIME_TMP_VAR = 0

    def setUp(self):
        self.server_url = "http://localhost/"
        self.odcs = ODCS(self.server_url)

    def test_wait_for_compose(self, get_compose, sleep):
        for state in ["done", "removed", "failed"]:
            get_compose.reset_mock()
            sleep.reset_mock()
            get_compose.side_effect = [
                {"state_name": "wait"},
                {"state_name": "generating"},
                {"state_name": state},
            ]
            self.odcs.wait_for_compose(1)

            self.assertEqual(sleep.mock_calls, [mock.call(1), mock.call(2)])
            self.assertEqual(get_compose.mock_calls, [mock.call(1)] * 3)

    @patch("time.time")
    def test_wait_for_compose_timeout(self, time_travel, get_compose, sleep):
        get_compose.side_effect = [{"state_name": "wait"}] * 2
        time_travel.side_effect = [1, 301]
        self.assertRaises(RuntimeError, self.odcs.wait_for_compose, 1)

    @patch("time.time")
    def test_wait_for_compose_elapsed_close_to_timeout(self, _time, get_compose, sleep):
        TestWaitForCompose._TIME_TMP_VAR = 0

        # Replace time.sleep() method which does not sleep, but updates the
        # _TIME_TMP_VAR with number of seconds it would sleep.
        def mocked_sleep(seconds):
            TestWaitForCompose._TIME_TMP_VAR += seconds

        sleep.side_effect = mocked_sleep

        # Replace time.time() method to return _TIME_TMP_VAR.
        def mocked_time():
            return TestWaitForCompose._TIME_TMP_VAR

        _time.side_effect = mocked_time

        get_compose.side_effect = [{"state_name": "wait"}] * 10
        self.assertRaises(RuntimeError, self.odcs.wait_for_compose, 1, 10)

        self.assertEqual(
            sleep.mock_calls, [mock.call(1), mock.call(2), mock.call(3), mock.call(4)]
        )

    @patch("odcs.client.odcs.ComposeLog.read")
    def test_wait_for_compose_watch_logs(self, log_read, get_compose, sleep):
        get_compose.side_effect = [
            {"state_name": "wait", "toplevel_url": "http://localhost/composes/odcs-1"},
            {
                "state_name": "generating",
                "toplevel_url": "http://localhost/composes/odcs-1",
            },
            {"state_name": "done", "toplevel_url": "http://localhost/composes/odcs-1"},
        ]
        log_read.side_effect = [None, "line\n"]
        self.odcs.wait_for_compose(1, watch_logs=True)

        self.assertEqual(sleep.mock_calls, [mock.call(10)])
        self.assertEqual(get_compose.mock_calls, [mock.call(1)] * 3)
        self.assertEqual(len(log_read.mock_calls), 2)


@patch("odcs.client.odcs.requests")
class TestComposeLog(unittest.TestCase):
    """Test ODCS.wait_for_compose"""

    def setUp(self):
        compose = {"toplevel_url": "http://localhost/composes/odcs-1"}
        self.compose_log = ComposeLog(compose)

    def test_compose_log_404(self, requests):
        requests.get.return_value.status_code = 404
        ret = self.compose_log.read()
        requests.get.assert_called_once_with(
            "http://localhost/composes/odcs-1/pungi-stderr.log",
            headers={"Range": "bytes=0-"},
        )
        self.assertEqual(ret, None)

    def test_compose_log_multiple_calls(self, requests):
        responses = [
            MagicMock(status_code=200, text="line\n"),
            MagicMock(status_code=200, text="another line\n"),
            MagicMock(status_code=416, text=""),
            MagicMock(status_code=200, text="another line\n"),
        ]
        requests.get.side_effect = responses
        length = 0
        for m in copy.copy(responses):
            ret = self.compose_log.read()
            self.assertEqual(ret, m.text)
            requests.get.assert_called_once_with(
                "http://localhost/composes/odcs-1/pungi-stderr.log",
                headers={"Range": "bytes=%d-" % length},
            )
            requests.get.reset_mock()
            length += len(ret)



# Copyright (c) 2020  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from mock import patch

from odcs.server import db
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_RESULTS
from odcs.server.pungi import PungiSourceType
from odcs.server.metrics import ComposesCollector, QueueLengthThread, WorkerCountThread
from .utils import ModelsBaseTest


class TestComposesCollector(ModelsBaseTest):
    def setUp(self):
        super(TestComposesCollector, self).setUp()
        self.collector = ComposesCollector()

    def test_composes_total(self):
        Compose.create(
            db.session,
            "unknown",
            PungiSourceType.MODULE,
            "testmodule:master",
            COMPOSE_RESULTS["repository"],
            60,
        )
        Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
        )
        db.session.commit()

        r = self.collector.composes_total()
        for sample in r.samples:
            if (
                sample.labels["source_type"] in ["module", "tag"]
                and sample.labels["state"] == "wait"
            ):
                self.assertEqual(sample.value, 1)
            else:
                self.assertEqual(sample.value, 0)

    def test_raw_config_composes_count(self):
        for i in range(15):
            Compose.create(
                db.session,
                "unknown",
                PungiSourceType.RAW_CONFIG,
                "foo#bar",
                COMPOSE_RESULTS["repository"],
                60,
            )
        for i in range(10):
            Compose.create(
                db.session,
                "me",
                PungiSourceType.RAW_CONFIG,
                "foo#hash%d" % i,
                COMPOSE_RESULTS["repository"],
                60,
            )
        db.session.commit()
        r = self.collector.raw_config_composes_count()
        for sample in r.samples:
            if sample.labels["source"] == "foo#bar":
                self.assertEqual(sample.value, 15)
            elif sample.labels["source"] == "foo#other_commits_or_branches":
                self.assertEqual(sample.value, 10)


class TestQueueLengthThread(ModelsBaseTest):
    @patch("odcs.server.metrics.celery_app")
    def test_update_metrics(self, celery_app):
        conn = celery_app.connection_or_acquire.return_value
        queue_declare = conn.default_channel.queue_declare
        queue_declare.return_value.message_count = 10

        thread = QueueLengthThread()
        thread.update_metrics()
        metrics = thread.queue_length.collect()
        for metric in metrics:
            queues = set()
            for sample in metric.samples:
                queues.add(sample.labels["queue_name"])
                self.assertEqual(sample.value, 10)
            self.assertEqual(
                queues, set(["cleanup", "pungi_composes", "pulp_composes"])
            )


class TestWorkerCountThread(ModelsBaseTest):
    @patch("odcs.server.metrics.celery_app")
    def test_update_metrics(self, celery_app):
        celery_app.control.ping.side_effect = [
            [
                {"worker-1@localhost": {"ok": "pong"}},
                {"worker-2@localhost": {"ok": "pong"}},
            ],
            [{"worker-1@localhost": {"ok": "pong"}}],
        ]

        # Both workers online.
        thread = WorkerCountThread()
        thread.update_metrics()
        metrics = thread.workers_total.collect()
        self.assertEqual(metrics[0].samples[0].value, 2)
        metrics = thread.workers.collect()
        for metric in metrics:
            for sample in metric.samples:
                self.assertEqual(sample.value, 1)

        # The worker-2 went offline.
        thread.update_metrics()
        metrics = thread.workers_total.collect()
        self.assertEqual(metrics[0].samples[0].value, 1)
        metrics = thread.workers.collect()
        for metric in metrics:
            for sample in metric.samples:
                if sample.labels["worker_name"] == "worker-1@localhost":
                    self.assertEqual(sample.value, 1)
                else:
                    self.assertEqual(sample.value, 0)



"""add lookaside_repos column

Revision ID: b2725d046624
Revises: 4514febd31fa
Create Date: 2019-01-26 09:17:33.392547

"""

# revision identifiers, used by Alembic.
revision = "b2725d046624"
down_revision = "4514febd31fa"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("lookaside_repos", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "lookaside_repos")



"""Add pungi_config_dump column.

Revision ID: 82172e6a3154
Revises: cd0781bbdab1
Create Date: 2020-02-05 13:20:59.014127

"""

# revision identifiers, used by Alembic.
revision = "82172e6a3154"
down_revision = "cd0781bbdab1"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("pungi_config_dump", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("composes", "pungi_config_dump")



"""Add multilib_arches and multilib_method

Revision ID: d1da07e15c54
Revises: f4bc999818d7
Create Date: 2018-09-03 14:06:40.565612

"""

# revision identifiers, used by Alembic.
revision = "d1da07e15c54"
down_revision = "f4bc999818d7"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("multilib_arches", sa.String(), nullable=True, default="")
    )
    op.add_column(
        "composes", sa.Column("multilib_method", sa.Integer(), nullable=True, default=0)
    )


def downgrade():
    op.drop_column("composes", "multilib_method")
    op.drop_column("composes", "multilib_arches")



# -*- coding: utf-8 -*-

# Copyright (c) 2020  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import time
import threading
from sqlalchemy import func
from prometheus_client import CollectorRegistry, Gauge
from prometheus_client.core import GaugeMetricFamily, CounterMetricFamily

from odcs.common.types import COMPOSE_STATES, PUNGI_SOURCE_TYPE_NAMES
from odcs.server.models import Compose
from odcs.server import log, conf, app

try:
    from odcs.server.celery_tasks import celery_app
    from celery.utils.objects import FallbackContext
    import amqp.exceptions

    CELERY_AVAILABLE = True
except ImportError:
    CELERY_AVAILABLE = False


registry = CollectorRegistry()


class ComposesCollector(object):
    def composes_total(self):
        """
        Returns `composes_total` GaugeMetricFamily with number of composes
        for each state and source_type.
        """
        counter = GaugeMetricFamily(
            "composes_total",
            "Total number of composes",
            labels=["source_type", "state"],
        )
        for state in COMPOSE_STATES:
            for source_type in PUNGI_SOURCE_TYPE_NAMES:
                count = Compose.query.filter(
                    Compose.source_type == PUNGI_SOURCE_TYPE_NAMES[source_type],
                    Compose.state == COMPOSE_STATES[state],
                ).count()

                counter.add_metric([source_type, state], count)
        return counter

    def raw_config_composes_count(self):
        """
        Returns `raw_config_composes_count` CounterMetricFamily with number of raw_config
        composes for each `Compose.source`. For raw_config composes, the Compose.source is
        stored in the `raw_config_key#commit_or_branch` format. If particular `Compose.source` is
        generated only few times (less than 5), it is grouped by the `raw_config_key` and
        particular `commit_or_branch` is replaced with "other_commits_or_branches" string.

        This is needed to handle the situation when particular raw_config compose is generated
        just once using particular commit hash (and not a branch name). These single composes
        are not that important in the metrics and therefore we group them like that.
        """
        counter = CounterMetricFamily(
            "raw_config_composes_count",
            "Total number of raw_config composes per source",
            labels=["source"],
        )
        composes = (
            Compose.query.with_entities(Compose.source, func.count(Compose.source))
            .filter(Compose.source_type == PUNGI_SOURCE_TYPE_NAMES["raw_config"])
            .group_by(Compose.source)
            .all()
        )

        sources = {}
        for source, count in composes:
            if count < 5:
                name = "%s#other_commits_or_branches" % source.split("#")[0]
                if name not in sources:
                    sources[name] = 0
                sources[name] += count
            else:
                sources[source] = count

        for source, count in sources.items():
            counter.add_metric([source], count)

        return counter

    def collect(self):
        yield self.composes_total()
        yield self.raw_config_composes_count()


class WorkerCountThread(threading.Thread):
    """
    Thread providing and updating following metrics:

    - celery_workers_totals - Number of alive workers.
    - celery_workers[worker_name] - 1 if worker is online, 0 if offline.
    """

    def __init__(self, registry=None):
        super(WorkerCountThread, self).__init__()
        self.daemon = True
        self.workers_total = Gauge(
            "celery_workers_totals", "Number of alive workers", registry=registry
        )
        self.workers_total.set(0)
        self.workers = Gauge(
            "celery_workers",
            "Number of alive workers",
            ["worker_name"],
            registry=registry,
        )
        self.worker_names = set()

    def update_metrics(self):
        log.info("[metrics] Getting number of workers.")
        try:
            celery_ping = celery_app.control.ping(timeout=15)
        except Exception:  # pragma: no cover
            log.exception("[metrics] Error pinging workers.")
            return

        # Set total number of workers.
        self.workers_total.set(len(celery_ping))

        # Set all known workers to 0 to mark them offline.
        for workers in celery_ping:
            self.worker_names |= set(workers.keys())
        for worker_name in self.worker_names:
            self.workers.labels(worker_name).set(0)

        # Set online workers to 1.
        for workers in celery_ping:
            for worker_name in workers.keys():
                self.workers.labels(worker_name).set(1)

    def run(self):
        while True:
            self.update_metrics()
            time.sleep(30)


class QueueLengthThread(threading.Thread):
    """
    Thread providing and updating following metrics:

    - celery_queue_length[queue_name] - Number of tasks waiting in the queue.
    """

    def __init__(self, registry=None):
        super(QueueLengthThread, self).__init__()
        self.daemon = True
        self.queue_length = Gauge(
            "celery_queue_length",
            "Number of tasks in the queue.",
            ["queue_name"],
            registry=registry,
        )

        # Get all the possible queue names from the config.
        self.queues = [conf.celery_cleanup_queue]
        for rules in conf.celery_router_config["routing_rules"].values():
            self.queues += rules.keys()
        # Initialize the queue length to 0.
        for queue in self.queues:
            self.queue_length.labels(queue).set(0)

        # Get the Celery connetion.
        self.connection = celery_app.connection_or_acquire()
        if isinstance(self.connection, FallbackContext):
            self.connection = self.connection.fallback()

    def update_metrics(self):
        for queue in self.queues:
            try:
                log.info("[metrics] Getting %s queue length." % queue)
                length = self.connection.default_channel.queue_declare(
                    queue=queue, passive=True
                ).message_count
                self.queue_length.labels(queue).set(length)
            except amqp.exceptions.ChannelError:
                # Queue not created yet.
                pass
            except Exception:  # pragma: no cover
                log.exception("[metrics] Error getting queue length.")

    def run(self):
        while True:
            self.update_metrics()
            time.sleep(30)


registry.register(ComposesCollector())


# Start the Celery metrics only on Frontend using the `before_first_request` decorator.
@app.before_first_request
def start_celery_metrics():
    if CELERY_AVAILABLE:
        # These threads are "daemonic". This means they are stopped automatically
        # by Python when main Python thread is stopped. There is no need to .join()
        # and since they are only updating metrics, they do not have to end up
        # gracefully and can just be "killed" by Python.
        worker_count_thread = WorkerCountThread(registry=registry)
        worker_count_thread.start()
        queue_length_thread = QueueLengthThread(registry=registry)
        queue_length_thread.start()



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import ssl
import os
import re
from celery import Celery
from six.moves.urllib.parse import urlparse
from datetime import datetime, timedelta

from odcs.server import conf, db, log
from odcs.server.backend import (
    generate_compose as backend_generate_compose,
    ComposerThread,
    RemoveExpiredComposesThread,
)
from odcs.server.utils import retry
from odcs.server.models import Compose, COMPOSE_STATES
from odcs.server.pungi import PungiSourceType


# Prepare the instances of classes with worker methods.
composer_thread = ComposerThread()
remove_expired_compose_thread = RemoveExpiredComposesThread()


# Create the Celery app.
if os.environ.get("ODCS_CELERY_BROKER_URL"):
    broker_url = os.environ["ODCS_CELERY_BROKER_URL"]
elif conf.celery_broker_url:
    broker_url = conf.celery_broker_url
else:
    broker_url = "amqp://localhost"

# Although the Celery handles amqps using Kombu Python module,
# it does not do that correctly.
#
# When using "amqps://" protocol, Kombu uses SSLTransport which effectively
# sets the `amqp.Connection.ssl = True`. This enables SSL, but it does it
# in a way when we cannot set any other SSL options like client cert/key
# or for example SNI hostname which needs to be used.
#
# This is done in this commit:
# In the https://github.com/celery/kombu/commit/701672c2ba732883869a42f061a10dbfc8fe2f30
#
# To workaround this, we detect "amqps" here and sets the `broker_use_ssl`. This
# allows us to set any SSL options like client cert/key or SNI header.
# The `broker_use_ssl` is directly passed to `amqp.Connection.ssl` which handles
# it.
#
# In the end, we change the `broker_url` protocol from "amqps" to "amqp", so
# the Celery wont' override the `amps.Connection.ssl` with `True`.
if broker_url.startswith("amqps://"):
    netloc = urlparse(broker_url).netloc
    host_port = netloc.split("@")[-1]
    host = host_port.split(":")[0]

    ssl_ctx = {}

    # Set the server_hostname only if it's configured in the CELERY_CONFIG,
    # otherwise Celery uses SNI everytime even if we don't need it.
    if "server_hostname" in conf.celery_config:
        broker_use_ssl = {
            "server_hostname": host,
            "context": {"purpose": ssl.Purpose.SERVER_AUTH},
        }
    else:
        broker_use_ssl = {}

    # Allow setting certificates using the CELERY_CONFIG.
    for key in ["ca_certs", "keyfile", "certfile"]:
        if key in conf.celery_config:
            broker_use_ssl[key] = conf.celery_config[key]

    conf.celery_config.update({"broker_use_ssl": broker_use_ssl})
    broker_url = broker_url.replace("amqps://", "amqp://")

celery_app = Celery("backend", broker=broker_url)
celery_app.conf.update(conf.celery_config)
celery_app.conf.update({"task_routes": ("odcs.server.celery_tasks.TaskRouter")})


class TaskRouter:
    """ Custom Celery router """

    def __init__(self):
        self.config = conf.celery_router_config

    def route_for_task(self, task_name, *args, **kwargs):
        """
        Method which celery expects to be defined on a custom router. Returns the payload
        with the queue selected for task
        """
        if task_name == self.config["cleanup_task"]:
            return {"queue": conf.celery_cleanup_queue}

        compose_id = args[0][0]
        compose = get_odcs_compose(compose_id)
        compose_md = compose.json()

        queue = self.__get_queue_for_compose(compose_md, task_name)

        return {"queue": queue}

    def __get_queue_for_compose(self, compose_md, task_name):
        """ Goes through routing rules configured for a task returns a queue on the first match. """
        rules = {}
        if self.config["routing_rules"].get(task_name):
            rules.update(self.config["routing_rules"][task_name])

        for queue, rule in rules.items():
            # if the rule has no properties its an automatic match.
            if rule:
                for key, value in rule.items():
                    if not compose_md.get(key):
                        raise ValueError(
                            (
                                "Task Router: Routing rule for queue %s for task %s contains an "
                                "invalid property: %s"
                            )
                            % (queue, task_name, key)
                        )

                    # if the value of the property from the rule and compose does not match, the
                    # whole rule is ignored and we go to the next rule
                    if isinstance(value, (tuple, list)):
                        values = value
                    else:
                        values = [value]

                    if not any(re.match(str(v), str(compose_md[key])) for v in values):
                        break
                else:
                    # if all of the properties and values match then the whole rule match
                    return queue
            else:
                # if the rule is emtpy its an wildcard and an automatic match
                return queue

        # if none of the rules applies the default queue is returned
        return self.config["default_queue"]


@celery_app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    # Add the cleanup task every 10 minutes. This can be hardcoded here.
    # It is just internal task to clean up the expired composes, mark
    # stuck composes as failed and so on.
    # This task also expires in 1 hour. This prevents situation when all backends
    # are broken for long time (hours/days) and queue of run_cleanup tasks grow
    # up and once backends come online, they are flooded with run_cleanup tasks.
    sender.add_periodic_task(10 * 60, run_cleanup.s(), expires=3600)


@retry(wait_on=RuntimeError)
def get_odcs_compose(compose_id):
    """
    Gets the compose from ODCS DB.
    """
    compose = Compose.query.filter(Compose.id == compose_id).first()
    if not compose:
        raise RuntimeError("No compose with id %d in ODCS DB." % compose_id)
    return compose


def generate_compose(compose_id):
    """
    Generates the compose with id `compose_id`.
    """
    compose = get_odcs_compose(compose_id)
    if compose.state != COMPOSE_STATES["wait"]:
        raise RuntimeError(
            "The 'generate_compose' called for compose not in 'wait' state: %r"
            % compose
        )
    compose.transition(COMPOSE_STATES["generating"], "Compose thread started")
    db.session.commit()
    backend_generate_compose(compose.id)


@celery_app.task
def generate_pungi_compose(compose_id):
    """
    Generates the Pungi based compose.
    """
    generate_compose(compose_id)


@celery_app.task
def generate_pulp_compose(compose_id):
    """
    Generates the Pungi based compose.
    """
    generate_compose(compose_id)


def get_current_celery_task_ids():
    """
    Returns a set with currently registered task IDs.
    """
    inspection = celery_app.control.inspect()

    # Active tasks are currently running tasks.
    active = []
    act_obj = inspection.active()
    if act_obj is not None:
        for i in act_obj.values():
            active += i
    active = [i["id"] for i in active]

    # Reserved tasks are assigned to particular worker, but
    # are not running yet.
    reserved = []
    res_obj = inspection.reserved()
    if res_obj is not None:
        for i in res_obj.values():
            reserved += i
    reserved = [i["id"] for i in reserved]

    return set(reserved + active)


def schedule_compose(compose):
    """
    Schedules the compose task and sets the `compose.celery_task_id`.
    Returns the updated `compose`.
    """
    if compose.source_type == PungiSourceType.PULP:
        result = generate_pulp_compose.delay(compose.id)
    else:
        result = generate_pungi_compose.delay(compose.id)
    compose.celery_task_id = result.id
    db.session.add(compose)
    db.session.commit()
    return compose


def reschedule_waiting_composes():
    """
    Gets all the composes stuck in the "wait" state and reschedules them.

    This method exists to pro-actively generate "wait" composes in case
    the Celery message from frontend to backend is lost from whatever reason.
    """
    # Composes transition from 'wait' to 'generating' quite fast.
    # The frontend changes the state of compose to 'wait', sends a message
    # to the bus and once some backend receives it, it moves it to
    # 'generating'. This should not take more than 3 minutes, so that's
    # the limit we will use to find out the stuck composes.
    # On the other hand, we don't want to regenerate composes older than
    # 3 days, because nobody is probably waiting for them. Just mark
    # them as "failed".
    now = datetime.utcnow()
    from_time = now - timedelta(days=3)
    to_time = now - timedelta(minutes=3)

    # Get composes which are in 'wait' state for too long.
    composes = (
        Compose.query.filter(
            Compose.state == COMPOSE_STATES["wait"], Compose.time_submitted < to_time
        )
        .order_by(Compose.id)
        .all()
    )

    # Get the current task ids registered by the workers.
    task_ids = get_current_celery_task_ids()

    for compose in composes:
        # Check if the task is already planned.
        if compose.celery_task_id in task_ids:
            continue

        if compose.time_submitted < from_time:
            compose.transition(
                COMPOSE_STATES["failed"],
                "Compose stuck in 'wait' state for longer than 3 days.",
            )
            continue

        log.info("%r: Rescheduling compose stuck in 'wait' state.", compose)
        schedule_compose(compose)


@celery_app.task
def run_cleanup():
    """
    Runs the cleanup.
    """
    remove_expired_compose_thread.do_work()
    composer_thread.fail_lost_generating_composes()
    reschedule_waiting_composes()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import os
import requests
import productmd.common

from odcs.server import conf


class PungiCompose(object):
    """Represents 3rd party Pungi Compose"""

    def __init__(self, variant_url):
        """
        Creates new PungiCompose.

        :param str metadata_url: URL to Pungi variant repository directory.
        """
        # The `variant_url` is for example "http://localhost/foo/compose/Server".
        self.variant_url = variant_url.rstrip("/")
        # The `variant_name` is for example `Server`
        self.variant_name = os.path.basename(self.variant_url)
        # The `metadata_url` is for example "http://localhost/foo/compose/metadata".
        self.metadata_url = os.path.join(os.path.dirname(self.variant_url), "metadata")

    def _fetch_json(self, url):
        """
        Fetches the json file represented by `url`.
        """
        r = requests.get(url, timeout=conf.net_timeout)
        r.raise_for_status()
        return r.json()

    def get_rpms_data(self):
        """
        Returns the data describing the RPMs in the pungi compose.
        :rtype: dict.
        :return: Dictionary with RPMs data in following format:
            {
                "sigkeys": set() with sigkeys used in the compose.
                "arches": set() with all the arches used in the compose.
                "builds": {
                    koji-build-nvr1: set() with the RPMs NEVRAs,
                    koji-build-nvr2: ...,
                    ...
                }
            }
        """
        ret = {}
        ret["sigkeys"] = set()
        ret["arches"] = set()
        ret["builds"] = {}

        # Fetch the rpms.json and get the part containing SRPMs
        # for the right variant.
        url = os.path.join(self.metadata_url, "rpms.json")
        data = self._fetch_json(url)
        srpms_per_arch = data.get("payload", {}).get("rpms", {}).get(self.variant_name)
        if not srpms_per_arch:
            raise ValueError(
                "The %s does not contain payload -> rpms -> %s "
                "section" % (url, self.variant_name)
            )

        # Go through the data and fill in the dict to return.
        for arch, srpms in srpms_per_arch.items():
            ret["arches"].add(arch)
            for srpm_nevra, rpms in srpms.items():
                packages = set()
                for rpm_nevra, rpm_data in rpms.items():
                    packages.add(rpm_nevra)
                    ret["sigkeys"].add(rpm_data["sigkey"])

                srpm_nvr = "{name}-{version}-{release}".format(
                    **productmd.common.parse_nvra(srpm_nevra)
                )
                ret["builds"][srpm_nvr] = packages

        return ret



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>


class PungiSourceType:
    KOJI_TAG = 1
    MODULE = 2
    REPO = 3
    PULP = 4
    RAW_CONFIG = 5
    BUILD = 6
    PUNGI_COMPOSE = 7


PUNGI_SOURCE_TYPE_NAMES = {
    "tag": PungiSourceType.KOJI_TAG,
    "module": PungiSourceType.MODULE,
    "repo": PungiSourceType.REPO,
    # This is not real PungiSourceType, but ODCS handles the PULP
    # as a extra source type by generating the .repo file pointing
    # to composes done by PULP/PUB.
    "pulp": PungiSourceType.PULP,
    # This allows to submit raw pungi config from predefined URLs in ODCS
    # server-side configuration.
    "raw_config": PungiSourceType.RAW_CONFIG,
    # Generates compose using exactly defined set of Koji builds without
    # pulling in RPMs from any Koji tag.
    "build": PungiSourceType.BUILD,
    # Generates compose using the same set of RPMs as in existing 3rd party
    # Pungi compose.
    "pungi_compose": PungiSourceType.PUNGI_COMPOSE,
}

INVERSE_PUNGI_SOURCE_TYPE_NAMES = {v: k for k, v in PUNGI_SOURCE_TYPE_NAMES.items()}

COMPOSE_STATES = {
    # Compose is waiting to be generated
    "wait": 0,
    # Compose is being generated.
    "generating": 1,
    # Compose is generated - done.
    "done": 2,
    # Compose has been removed.
    "removed": 3,
    # Compose generation has failed.
    "failed": 4,
}

INVERSE_COMPOSE_STATES = {v: k for k, v in COMPOSE_STATES.items()}

MULTILIB_METHODS = {
    "none": 0,
    # Packages whose name ends with "-devel" or "-static" suffix will
    # be considered as multilib.
    "runtime": 1,
    # Packages that install some shared object file "*.so.*" will be
    # considered as multilib.
    "devel": 2,
    # All pakages will be considered as multilib.
    "all": 4,
}

INVERSE_MULTILIB_METHODS = {v: k for k, v in MULTILIB_METHODS.items()}

COMPOSE_RESULTS = {
    "repository": 1,
    "iso": 2,
    "ostree": 4,
    "boot.iso": 8,
}

COMPOSE_FLAGS = {
    "no_flags": 0,
    # Compose without pulling-in the dependencies of packages or modules
    # defined for a compose.
    "no_deps": 1,
    # Compose without pulling-in packages from the parent Koji tags.
    "no_inheritance": 2,
    # For "pulp" source_type, include unpublished Pulp repos.
    "include_unpublished_pulp_repos": 4,
    # Abort the compose when some package has broken dependencies.
    "check_deps": 8,
    # Include also module builds in the "done" state.
    "include_done_modules": 16,
    # For "pulp" source_type, ignore any repos do not exist in the remote Pulp
    # instance.
    "ignore_absent_pulp_repos": 32,
    # Do not reuse old composes.
    "no_reuse": 64,
}

INVERSE_COMPOSE_FLAGS = {v: k for k, v in COMPOSE_FLAGS.items()}



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import six
import unittest

from mock import patch

from odcs.server import conf
from odcs.server import db
from odcs.server.models import Compose
from odcs.server.pungi import PungiSourceType
from odcs.common.types import COMPOSE_RESULTS
from odcs.common.types import COMPOSE_STATES
from .utils import ModelsBaseTest

try:
    import rhmsg
except ImportError:
    rhmsg = None

try:
    import fedora_messaging
except ImportError:
    fedora_messaging = None


@unittest.skipUnless(rhmsg, "rhmsg is required to run this test case.")
@unittest.skipIf(six.PY3, "rhmsg has no Python 3 package so far.")
class TestRHMsgSendMessageWhenComposeIsCreated(ModelsBaseTest):
    """Test send message when compose is created"""

    disable_event_handlers = False

    def setUp(self):
        super(TestRHMsgSendMessageWhenComposeIsCreated, self).setUp()

        # Real lock is not required for running tests
        self.mock_lock = patch("threading.Lock")
        self.mock_lock.start()

    def tearDown(self):
        self.mock_lock.stop()

    def setup_composes(self):
        self.compose = Compose.create(
            db.session,
            "mine",
            PungiSourceType.KOJI_TAG,
            "f25",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

    @patch.object(conf, "messaging_backend", new="rhmsg")
    @patch("rhmsg.activemq.producer.AMQProducer")
    @patch("proton.Message")
    def assert_messaging(self, compose, Message, AMQProducer):
        db.session.commit()

        self.assertEqual(
            json.dumps({"event": "state-changed", "compose": compose.json()}),
            Message.return_value.body,
        )

        producer_send = AMQProducer.return_value.__enter__.return_value.send
        producer_send.assert_called_once_with(Message.return_value)

    def test_send_message(self):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )

        self.assert_messaging(compose)

    def test_message_on_state_change(self):
        compose = (
            db.session.query(Compose).filter(Compose.id == self.compose.id).all()[0]
        )
        compose.state = COMPOSE_STATES["generating"]

        self.assert_messaging(compose)


@unittest.skipUnless(
    fedora_messaging, "fedora_messaging is required to run this test case."
)
class TestFedoraMessagingSendMessageWhenComposeIsCreated(ModelsBaseTest):
    """Test send message when compose is created"""

    disable_event_handlers = False

    def setUp(self):
        super(TestFedoraMessagingSendMessageWhenComposeIsCreated, self).setUp()

        # Real lock is not required for running tests
        self.mock_lock = patch("threading.Lock")
        self.mock_lock.start()

    def tearDown(self):
        self.mock_lock.stop()

    def setup_composes(self):
        self.compose = Compose.create(
            db.session,
            "mine",
            PungiSourceType.KOJI_TAG,
            "f25",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

    @patch.object(conf, "messaging_backend", new="fedora-messaging")
    @patch("fedora_messaging.api.Message")
    @patch("fedora_messaging.api.publish")
    def assert_messaging(self, compose, publish, Message):
        # The db.session.commit() calls on-commit handler which produces the fedora-messaging
        # message.
        db.session.commit()

        Message.assert_called_once_with(
            topic="odcs.compose.state-changed",
            body={"event": "state-changed", "compose": compose.json()},
        )

        publish.assert_called_once_with(Message.return_value)

    def test_send_message(self):
        compose = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            3600,
        )

        self.assert_messaging(compose)

    def test_message_on_state_change(self):
        compose = (
            db.session.query(Compose).filter(Compose.id == self.compose.id).all()[0]
        )
        compose.state = COMPOSE_STATES["generating"]

        self.assert_messaging(compose)



"""Add parent_pungi_compose_ids.

Revision ID: 7b143656694f
Revises: 512890e6864d
Create Date: 2020-08-17 08:07:45.861953

"""

# revision identifiers, used by Alembic.
revision = "7b143656694f"
down_revision = "512890e6864d"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("parent_pungi_compose_ids", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("composes", "parent_pungi_compose_ids")
